/************************************************************************
	POP.C
	
	A simple implementatio of a client that retrive mail from a POP3 server.
	Written by Davide Bianchi (davide_bianchi@usa.net)
	
	Version 1.00 - September 1999
	
	Based on the RFC 1725
	http://info.internet.edu/in-notes/rfc/files/rfc1725.txt

	Revision:
		October 1999 added the support for Windows Socket (winsock).


	This program is freely available and redistributable under the GPL
	licensing.
	
***********************************************************************/

#ifdef _WIN_
	#include <stdlib.h>
	#include <sys/types.h>
	#include <winsock.h>
#else
	#include <unistd.h>
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <netdb.h>
#endif				

#include <string.h>
#include <sys/stat.h>
#include <stdio.h>

#define POP_PORT	110
#define BUFLEN		4096

int sock;
char *buffer;

/**********************************************************
This is required because the idiot Windows uses 
closesocket() instead of the standard close().
**********************************************************/
#ifdef _WIN_

void close( int sock )
{
	closesocket( sock );
	return;
}

#endif

/***********************************************************
	SOME USEFULL FUNCTION
***********************************************************/

/***********************************************************
	atoaddr()
	Convert a string to an Internet address, the address
	can be sent as an IP address xxx.xxx.xxx.xxx or like
	an hostname.

	Return the In_Addr or NULL if something wrong happen
***********************************************************/
struct in_addr *atoaddr( char* address )
{
  struct hostent *host;
  static struct in_addr saddr;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if (saddr.s_addr != -1)
  {
    return &saddr;
  }
  
  /* try to locate the IP address from the name */
  host = gethostbyname( address );
  if (host != NULL)
  {
    return (struct in_addr *) *host->h_addr_list;
  }
  
  /** GEEE! Unable to locate the address */
  return NULL;
}

/***********************************************************
	Send
	Send data into the socket, this is used to coop with
	the fact that the Socket can receive less data than
	we want to send...	

	If something goes wrong, the function return -1
************************************************************/
int Send( int sock, char *buffer, int len, int flags )
{
	int count = 0;
	int res;

	while( count < len )
	{
  	res = send( sock, buffer + count, len - count, flags );
    if( res <= 0 )
		{
			perror( "Error during the Send" );
    	return -1;
		}
    else
    	count += res;
	}

	/* return the number of sent data */
	return count;
	
}

/******************************************************
	Receive()
	Handle the "receive" process trought the socket,
  when the buffer is received, the last char is
	terminate with a leading \0.

	If something goes wrong, the function return -1
	else the number of character received is returned.
******************************************************/
int Receive( int sock, char *buffer, int buffsize, int flags )
{
	char *p;
	int	res = recv( sock, buffer, buffsize, flags );

	if( res <= 0 )
	{
		perror("Error during the recv()" );
		return -1;
	}
	else
    *( buffer + res ) = '\0';

	return res;
	
}

/*****************************************************
	checkreply
	This function is used to check the result of an
	operation. The result could be +OK or something
	diffrent. Only the +OK is considered as a "OK"

	If OK is received the function return 1, else
	a 0 is returned.	
*****************************************************/
int checkreply( char *buffer )
{
	char *ok = "+OK";
	
	if( memcmp( buffer, ok, 3 ) == 0 )
	{
		return 1;
	}
	else
	{
		perror( buffer );
		return 0;
	}
}

/*****************************************************
	computemail
	
	This function is used to check how much mail are
	in the POP.
	warning: absolutely no error control....
	The function return the number of mail to retrive
	or 0.
*****************************************************/

int computemail( char *buffer )
{
	int mailnum;
	char *tok = strtok( buffer, " " );

	tok = strtok( NULL, " " );
	
	mailnum = atoi( tok );
	return(mailnum);
	
}

/*****************************************************
	checklastblock()
	check if the buffer contains the last-block-mark
	for the last block of a single message.
	The last block is marked with a "." alone in one
	row.
	The function return 1 if the marker is found.
*****************************************************/
int checklastblock( char *buffer )
{
	char *tok=strtok( buffer, "\n");
	while( tok )
	{
		if( memcmp( tok, ".",1 ) == 0 )
			return(1);
		tok=strtok( NULL, "\n" );
	}
	return(0);
}

/*****************************************************
	sendmessage()
	This simple function is used to send a message to
	the server and get the corresponding reply.
*****************************************************/
int sendmessage( char* message )
{
	int i;

	sprintf( buffer, "%s\n\0", message );
	i=Send( sock, buffer, strlen( buffer ), 0 );
	if( i <= 0 )
	{
		perror("Error during sent");
		return(-1);
	}

	/* receive ack */
	i=Receive( sock, buffer, BUFLEN, 0 );
	if( i <= 0 || ! checkreply( buffer ) )
	{
		perror( "Error during receive" );
		return(-1);
	}

	return(0);
}

/*****************************************************
	closeall()
	This function is used to close everything and
	remove the allocated memory.
*****************************************************/
void closeall()
{
	close(sock);
	if( buffer != NULL )
		free( buffer );
	return;
}

/*****************************************************
	Main program
*****************************************************/
int main(int argc, char *argv[])
{
  struct sockaddr_in server;  				/* server address info */
  struct in_addr *addr; 							/* Internet address of server */
  
  /* used to keep track of status */
  int i;
  
  /* counter of mails */
  int count;
  int totmail;
  
  /* file to write the mail into */
	FILE *f;

	/* 
		 Under Windows, you have to call the WSAStartup to open
	   a socket
	*/
	#ifdef _WIN_
		WSADATA WSAData;
	  WSAStartup(MAKEWORD(1,1), &WSAData);
	#endif
	
	/* check if the user passed arguments... */
  if (argc <3 ) {
    printf( "Usage: %s <pop_host> <username> <password>\n", argv[0] );
    return(0);
  }

	/* try to locate the server */
  addr = atoaddr(argv[1]);
  if( addr == NULL )
  {
  	perror("Unknown host." );
  	return(1);
  }
  
	/* build the socket */
  sock = socket( AF_INET, SOCK_STREAM, 0 );
  if (sock < 0) {
    perror("Socket() failed!\n");
    return(1);
  }

  /* buffer used to read the information */
  buffer = (char*) malloc( BUFLEN );
	if( buffer == NULL )
	{
		/* error allocating memory */
		perror("Memory error: could not allocate memory!\n" );
		return(1);
	}

  /* Set up the server address */
  memset((char *) &server, 0, sizeof(server));
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = addr->s_addr;
  server.sin_port = htons( POP_PORT );

	/*
		connect to the host
	*/
	if( connect( sock, (struct sockaddr*) &server, sizeof( server ) ) )
	{
		perror("Error during connect\n" );
		closeall();
		return(1);
	}

	/*
		receive the ACK from the server, should be "+OK"
	*/
	i=Receive( sock, buffer, BUFLEN, 0 );
	if( i < 0 || ! checkreply( buffer ) )
	{
		/* error during the receive */
		closeall();
		return(1);
	}
	
	/* send the Username */
	sprintf( buffer, "USER %s", argv[2] );
	if(	sendmessage( buffer ) )
	{
		closeall();
		return(1);
	}

	/* 
		send the Password
	
		WARNING!!! THIS IS A SERIOUS SECURITY HOLE!
		The password is sent like PLAIN-TEXT.
		
	*/
	sprintf( buffer, "PASS %s", argv[3] );
	if( sendmessage( buffer ) )
	{
		closeall();
		return(1);
	}

	/* good, send the STAT command to retrive the number of mail */
	sprintf( buffer, "STAT" );
	if( sendmessage( buffer ) )
	{
		closeall();
		return(1);
	}

	/* compute how much mail I have to retrive */
	totmail = computemail( buffer );
	for( count=1; count <= totmail; count++ )
	{

		/* prepare the filename to save the mail */
		sprintf( buffer, "%s%d", argv[2], count );
		
		/* open the file */
		f = fopen( buffer, "w+");
		if( f==NULL )
		{
			perror("Unable to save the mail\n");
			closeall();
			return(1);
		}

		/* retrive the i-mail... */
		sprintf( buffer, "RETR %d\n\0", count );
		if( sendmessage( buffer ) )
		{
			closeall();
			return(1);
		}

		/* get the message  */
		while( 1 )
		{
			/* read a bunch of data */
			Receive( sock, buffer, BUFLEN, 0 );

			/* 
				for compatibility, I have to change the 13/10 pairs
				into \n char...
			*/
			for( i=0; i<strlen( buffer ); i++ )
			{
				if( buffer[i] == 10 )
				{
					buffer[i-1]=' ';
					buffer[i]='\n';
				}
			}

			/* save the line in the file */
			fprintf(f, "%s\n", buffer );

			/* check if the block is the last for this msg */
			if( checklastblock( buffer ))
				break;
		}

		/* close the file */
		fclose( f );

		/* remove the mail */
		sprintf( buffer, "DELE %d\n\0", count );
		if( sendmessage( buffer ) )
		{
			closeall();
			return(1);
		}
	}
	
	/* send the QUIT command */
  sendmessage( "QUIT" );
	closeall();
  return 0;
  
}
